"""
reactionmenu • discord.py pagination
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A package to create a discord.py paginator. Supports pagination with Discords Buttons feature and reactions.

:copyright: (c) 2021 Defxult#8269
:license: MIT

"""

from .buttons import Button, ButtonType, ComponentsButton
from .buttons_menu import ButtonsMenu
from .core import ReactionMenu
from .text import TextMenu

__version__ = '1.1.0'
__source__ = 'https://github.com/Defxult/reactionmenu'
